/* 

File: main.cpp

Copyright (c) 2010 Department of Information Engineering, University of Padova, Italy
Contributors: Davide Zordan, Marco Vison, 11 Feb. 2010
All rights reserved

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.
    * The name of the author may not be used to endorse or promote
      products derived from this software without specific prior
      written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "main.h"
#include "nesta.h"

/*
 * Prints an error message on the standard output
 */
void Die(char *mess) {
	perror(mess);
	printf("%s\n", mess);
	exit(EXIT_FAILURE);
}

/*
 * Reads data in the buffer of length dataLen.
 * Data are stored in the buffer separated by a string terminator char.
 * The  function switches on the commands received such as
 * PREAMBLE, NEW_MEASURE, END_OF_SENSORS, and behaves
 * consequently.
 */
int readBuffer(char* buffer, int dataLen) {
	int retVal = 0;
	char* bufPos;
	int64_t timestamp;
	int16_t sensorId;
	float value;
//	printf("Starting buffer reading...");
	bufPos = &buffer[0];
	while (bufPos < &buffer[0] + dataLen) {
		int control = atoi(bufPos);
		bufPos += strlen(buffer) + 1;
		switch (control) {
		case PREAMBLE:
			numSensors = (int32_t) atoi(bufPos);
			bufPos += strlen(bufPos) + 1;
			numMeasures = (int32_t) atoi(bufPos);
			bufPos += strlen(bufPos) + 1;
			ToM = (int32_t) atoi(bufPos);
			bufPos += strlen(bufPos) + 1;
			printf("numsensors: %d\n", numSensors);

			//Inizializzazione struttura dati contentente le misurazioni
			initMeasures = (sensorMeasures*) malloc(sizeof(sensorMeasures)
					* numSensors);
			finalMeasures = (sensorMeasures*) malloc(sizeof(sensorMeasures)
					* numSensors);
			for (int i = 0; i < numSensors; i++) {
				initMeasures[i].value = new float[numMeasures];
				*initMeasures[i].value = -1.0;
				initMeasures[i].valuePos = &initMeasures[i].value[0];
				initMeasures[i].timestamp = new int64_t[numMeasures];
				initMeasures[i].tsPos = &initMeasures[i].timestamp[0];

				finalMeasures[i].value = new float[numMeasures];
				finalMeasures[i].valuePos = &initMeasures[i].value[0];
				finalMeasures[i].timestamp = new int64_t[numMeasures];
				finalMeasures[i].tsPos = &initMeasures[i].timestamp[0];
			}
			retVal = PREAMBLE;
			break;
		case NEW_MEASURE:
			sensorId = (int16_t) atoi(bufPos);
			bufPos += strlen(bufPos) + 1;
			value = (float) atof(bufPos);
			bufPos += strlen(bufPos) + 1;
			timestamp = (int64_t) atol(bufPos);
			bufPos += strlen(bufPos) + 1;

			initMeasures[sensorId - 1].id = sensorId;
			*initMeasures[sensorId - 1].valuePos = value;
			initMeasures[sensorId - 1].valuePos++;
			*initMeasures[sensorId - 1].tsPos = timestamp;
			initMeasures[sensorId - 1].tsPos++;
			/*
                         * All the values are also copied into finalMeasures
			 */
			finalMeasures[sensorId - 1].id = sensorId;
			*finalMeasures[sensorId - 1].valuePos = value;
			finalMeasures[sensorId - 1].valuePos++;
			*finalMeasures[sensorId - 1].tsPos = timestamp;
			finalMeasures[sensorId - 1].tsPos++;

			retVal = NEW_MEASURE;
			break;
		case END_OF_SENSORS:
			return END_OF_SENSORS;
		default:
			retVal = -1;
			break;
		}
	}
	return retVal;
}

/*
 * Allocates the buffer and receives new data from socket sock
 * while there are new data to read. Then all the separator chars '|'
 * in the buffer are substituted by the string terminator char,
 * before passing it to the readBuffer function
 */
int receiveData(int sock) {
    char* buffer = new char[BUFFSIZE];
    int received = -1;

    do {
        memset(buffer, 0, BUFFSIZE);
        if ((received = recv(sock, buffer, BUFFSIZE, 0)) < 0) {
            Die("Failed to receive additional bytes from client");
            return -1;
        }
//        printf("datalen: %d\n", received);

        for (int i = 0; i < received; i++) {
//            cout << buffer[i];
            if (buffer[i] == '|') {   // Look for a |
                buffer[i] = '\0';     // replace with a \0
            }
        }
        if (readBuffer(buffer, received) == END_OF_SENSORS)
            break;
    } while (received > 0);
    return 0;
}

/*
 * Reconstructs the data in structure initMeasures.
 * First gets the most recent timestamp in the variable mostRecent.
 * Then build a training set, and reconstruct the signal through
 * the NESTA algorithm.
 */
void reconstruct() {
	printf("Starting reconstruction...\n");

	time_t mostRecent = 0;
	int conta = 0; //counts the number of element of the most recent measure

	for (int i = 0; i < numSensors; i++) {
		if ((time_t) initMeasures[i].timestamp[0] > mostRecent) {
			mostRecent = (time_t) initMeasures[i].timestamp[0];
			conta = 1;
		} else if ((time_t) initMeasures[i].timestamp[0] == mostRecent) {
			conta++;
		}
	}
	printf("Most recent timestamp: %ld\n", mostRecent);

	ColumnVector y(conta);
	Matrix Phi(conta, numSensors);
	Phi = 0;
	ColumnVector x(numSensors);

	if (last[ToM].depth == 0) {
		// first round
            for (int i = 0; i < numSensors; i++) {
			y(i + 1) = initMeasures[i].value[0];
			x(i + 1) = initMeasures[i].value[0];
		}
	} else if ((last[ToM].depth == 1) || (conta == numSensors)) {
		// second round or previous iteration ptx = 1
		for (int i = 0; i < numSensors; i++) {
			y(i + 1) = initMeasures[i].value[0];
			x(i + 1) = initMeasures[i].value[0];
			Phi(i + 1, i + 1) = 1;
			
		}
                
                delete last[ToM].Phi_prec;
		delete last[ToM].x_prec;
		delete last[ToM].y_prec;

		// Update of 'last'
                last[ToM].p_tx = last[ToM].p_tx - C2 / numSensors;
		last[ToM].Phi_prec = new Matrix(conta, numSensors);
		last[ToM].x_prec = new ColumnVector(numSensors);
		last[ToM].y_prec = new ColumnVector(conta);

		*last[ToM].Phi_prec = Phi;
		*last[ToM].x_prec = x;
		*last[ToM].y_prec = y;

	} else {
		// Initialization of y, Training Set and Phi
		Matrix Tr_Set(numSensors, numMeasures - 1);

		int k = 1;
		for (int i = 0; i < numSensors; i++) {
			initMeasures[i].resetPtrs();
			finalMeasures[i].resetPtrs();

			if (initMeasures[i].timestamp[0] == mostRecent) {
                                // must put it in y and write a 1 in Phi
				y(k) = initMeasures[i].value[0];
				Phi(k, i + 1) = 1;
				k++;
				// further measurements should be included in the training set
				for (int j = 1; j < numMeasures; j++) {
					Tr_Set(i + 1, j) = initMeasures[i].value[j];
				}
			} else {
				for (int j = 0; j < numMeasures - 1; j++) {
					Tr_Set(i + 1, j + 1) = initMeasures[i].value[j];
				}
			}
		}

		int Tr_Set_Depth = 0;
		for (int j = 0; j < numMeasures - 1; j++) {
			if (Tr_Set.Column(j+1).Sum() > 0) {
				Tr_Set_Depth++;
			}
		}

		Matrix Resized_Tr_Set(numSensors, Tr_Set_Depth);
		for (int j = 1; j <= Tr_Set_Depth; j++) {
			Resized_Tr_Set.Column(j) = Tr_Set.Column(j);
		}

		// ready to procede with PCA and Nesterov
		ColumnVector m(numSensors);
		Matrix Psi(numSensors, numSensors);

		PCA_Transform_Matrix(Resized_Tr_Set, Psi, m);

		ColumnVector y1(conta);
		y1 << y - Phi * m;

		ColumnVector s(numSensors);
		s << Nesterov(Phi * Psi, y1);


		x = Psi * s + m; //signal recovery

		double error = sqrt(((y & *last[ToM].y_prec) - (*last[ToM].Phi_prec
				* *last[ToM].x_prec & Phi * x)).SumSquare()) / sqrt((y
				& *last[ToM].y_prec).SumSquare());


		if (error > err_threshold) {
			C1 * last[ToM].p_tx > 1.0 ? last[ToM].p_tx = 1 : last[ToM].p_tx
					= C1 * last[ToM].p_tx;
		} else {
			last[ToM].p_tx - C2 / numSensors > p_tx_min ? last[ToM].p_tx
					= last[ToM].p_tx - C2 / numSensors : last[ToM].p_tx
					= p_tx_min;
		}

		// Update of  'last'
		delete last[ToM].Phi_prec;
		delete last[ToM].x_prec;
		delete last[ToM].y_prec;

		last[ToM].Phi_prec = new Matrix(conta, numSensors);
		last[ToM].x_prec = new ColumnVector(numSensors);
		last[ToM].y_prec = new ColumnVector(conta);

		*last[ToM].Phi_prec = Phi;
		*last[ToM].x_prec = x;
		*last[ToM].y_prec = y;
	}
	last[ToM].depth++;

	for (int i = 0; i < numSensors; i++) {
		initMeasures[i].resetPtrs();
		finalMeasures[i].resetPtrs();
		if (initMeasures[i].timestamp[0] != mostRecent) {
			printf("Entrato nell'if per il sensore con id: %d\n",
					initMeasures[i].id);
			int j;
			for (j = 0; j < numMeasures - 1; j++) {
				finalMeasures[i].id = initMeasures[i].id;
				finalMeasures[i].value[j + 1] = initMeasures[i].value[j];
				finalMeasures[i].timestamp[j + 1]
						= initMeasures[i].timestamp[j];
			}

			finalMeasures[i].id = initMeasures[i].id;
			finalMeasures[i].value[0] = x(i + 1);
			finalMeasures[i].timestamp[0] = mostRecent; //initMeasures[i].timestamp[0];
		} else {
			for (int j = 0; j < numMeasures - 1; j++) {
				finalMeasures[i].id = initMeasures[i].id;
				finalMeasures[i].value[j] = initMeasures[i].value[j];
				finalMeasures[i].timestamp[j] = initMeasures[i].timestamp[j];
			}
			finalMeasures[i].id = initMeasures[i].id;
			finalMeasures[i].value[0] = x(i + 1);
			finalMeasures[i].timestamp[0] = mostRecent;
		}
	}
}

/*
 * Returns the finalMeasures data encoded as chars separated by '|'
 * to the web application
 */
int sendResults(int sock) {
	//printf("Restituzione dati ricostruiti...\n");
	char* buffer = new char[BUFFSIZE];
	memset(buffer, 0, BUFFSIZE);
	sprintf(buffer, "%d|%f|", RESPONSE,last[ToM].p_tx);
        printf("%d|%f|", RESPONSE,last[ToM].p_tx);
	int dataLen = strlen(buffer);

	if (send(sock, buffer, dataLen, 0) != dataLen) {
		Die("Failed to send bytes to client");
		return -1;
	}

	//For each sensor
	for (int i = 0; i < numSensors; i++) {
		memset(buffer, 0, BUFFSIZE);
		printf("%d|%d|%f|%ld|", NEW_MEASURE, finalMeasures[i].id,
				finalMeasures[i].value[0], finalMeasures[i].timestamp[0]);
		sprintf(buffer, "%d|%d|%f|%ld|", NEW_MEASURE, finalMeasures[i].id,
				finalMeasures[i].value[0], finalMeasures[i].timestamp[0]);
		int dataLen = strlen(buffer);

		if (send(sock, buffer, dataLen, 0) != dataLen) {
			Die("Failed to send bytes to client");
			return -1;
		}
	}
        printf("%d|",END_OF_SENSORS);
	sprintf(buffer, "%d|", END_OF_SENSORS);
	if (send(sock, buffer, dataLen, 0) != dataLen) {
		Die("Failed to send bytes to client");
		return -1;
	}
	close(sock);
	return 0;
}

/*
 * Application's starting point
 */
int main(int argc, char *argv[]) {
	initMeasures = (sensorMeasures*) malloc(sizeof(sensorMeasures));
	fprintf(stdout, "Starting...\n");

	
	int serversock, clientsock;
	struct sockaddr_in server, client;

	if (argc != 2) {
		fprintf(stderr, "USAGE: echoserver <port>\n");
		exit(1);
	}

	/* Create the TCP socket */
	if ((serversock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
		Die("Failed to create socket");
	}

	/* Construct the server sockaddr_in structure */
	memset(&server, 0, sizeof(server)); /* Clear struct */
	server.sin_family = AF_INET; /* Internet/IP */
	server.sin_addr.s_addr = htonl(INADDR_ANY); /* Incoming addr */
	server.sin_port = htons(atoi(argv[1])); /* server port */
	/* Bind the server socket */
	if (bind(serversock, (struct sockaddr *) &server, sizeof(server)) < 0) {
		Die("Failed to bind the server socket");
	} else {
		fprintf(stdout, "Bind successful...\n");
	}
	/* Listen on the server socket */
	if (listen(serversock, MAXPENDING) < 0) {
		Die("Failed to listen on server socket");
	} else {
		fprintf(stdout, "Listen successful...\n");
	}
	/*
	 * Initialization of  "last"
	 */
	for (int i = 0; i < 5; i++) {
		last[i].Phi_prec = new Matrix(1, 1);
		(*last[i].Phi_prec)(1, 1) = 0;
		last[i].x_prec = new ColumnVector(1);
		(*last[i].x_prec)(1) = 0;
		last[i].y_prec = new ColumnVector(1);
		(*last[i].y_prec)(1) = 0;
		last[i].p_tx = 1.0;
		last[i].depth = 0;
	}
	/* Run until cancelled */
	while (1) {
		unsigned int clientlen = sizeof(client);
		printf("Wait for client connection...\n");
		if ((clientsock = accept(serversock, (struct sockaddr *) &client,
				&clientlen)) < 0) {
			Die("Failed to accept client connection");
		}
		fprintf(stdout, "Client connected: %s:%d\n",
				inet_ntoa(client.sin_addr), ntohs(client.sin_port));

		now = time(NULL);
		receiveData(clientsock); // receive socket data
		reconstruct(); // reconstruct received data
		sendResults(clientsock); // return reconstructed data to web application
		printf("All operations successfully completed...\n");
	}
	free(initMeasures);
	exit(EXIT_SUCCESS);
}

/*
 * Compares two timestamps and returns the greater one
 */
time_t max(time_t a, time_t b) {
	if (a >= b)
		return a;
	else
		return b;
}

